#ifndef _HEX_EDIT_STATE_H_
#define _HEX_EDIT_STATE_H_

#include <kapp.h>
#include <qfont.h> 


struct SDisplayLine
{
  enum EViewMode
  {
    hexadecimal = 0,
    octal,
    binary,
    textOnly,
  };

  SDisplayLine( void )
  {
    lineSize[hexadecimal] = 16;
    lineSize[octal]       = 16;
    lineSize[binary]      = 8;
    lineSize[textOnly]    = 64;

    columnSize[hexadecimal] = 1;
    columnSize[octal]       = 1;
    columnSize[binary]      = 1;
    columnSize[textOnly]    = 64;
  }


  uint lineSize[4];
  uint columnSize[4];
};

struct SDisplayLayout
{
  enum EViewMode
  {
    hexadecimal = 0,
    octal,
    binary,
    textOnly,
    decimal,
    hide,
  };

  SDisplayLayout( void )
  {
    offsetMode       = hexadecimal;
    primaryMode      = hexadecimal;
    secondaryMode    = textOnly;
    offsetUpperCase  = false;
    primaryUpperCase = false;
    offsetVisible    = true;
    lineSize         = 16;
    columnSize       = 1;
    lockLine         = false;
    lockColumn       = true;
    columnSpacing    = -1;
    separatorMarginWidth = 5;
    edgeMarginWidth      = 5;
    leftSeparatorWidth   = 1;
    rightSeparatorWidth  = 1;
  }

  void verify( void )
  {
    if( lineSize < 1 ) { lineSize  = 1; }
    if( columnSize < 1 ) { columnSize = 1; }
    if( columnSize > lineSize ) { columnSize = lineSize; }
    if( columnSpacing == 0 ) { columnSize = lineSize; }
    if( primaryMode == textOnly ) { secondaryMode = hide; columnSpacing = 0; }
  }

  bool showSecondary( void )
  {
    if( primaryMode == textOnly || secondaryMode == hide )
    {
      return( false );
    }
    else
    {
      return( true );
    }
  }
  
  const char *modeStrings( uint index )
  {
    static char *str[6] = 
    {
      "hexadecimal",
      "octal",
      "binary",
      "textOnly",
      "decimal",
      "hide",
    };
    return( str[index] );
  }


  const char *primaryModeString( void )
  {
    return( modeStrings( primaryMode > textOnly ? hexadecimal : primaryMode ));
  }

  const char *secondaryModeString( void )
  {
    return( modeStrings( secondaryMode == textOnly ? textOnly : hide ) );
  }

  const char *offsetModeString( void )
  {
    return( modeStrings( offsetMode == hexadecimal ? hexadecimal : hide ) );
  }


  void setPrimaryMode( const char *str )
  {
    if( str == 0 || strcmp( str, "hexadecimal" ) == 0 )
    {
      primaryMode = hexadecimal;
    }
    else if( strcmp( str, "octal" ) == 0 )
    {
      primaryMode = octal;
    }
    else if( strcmp( str, "binary" ) == 0 )
    {
      primaryMode = binary;
    }
    else if( strcmp( str, "textOnly" ) == 0 )
    {
      primaryMode = textOnly;
    }
    else
    {
      primaryMode = hexadecimal;
    }
  }

  void setSecondaryMode( const char *str )
  {
    if( str == 0 || strcmp( str, "textOnly" ) == 0 )
    {
      secondaryMode = textOnly;
    }
    else
    {
      secondaryMode = hide;
    }
  }

  void setOffsetMode( const char *str )
  {
    if( str == 0 || strcmp( str, "hexadecimal" ) == 0 )
    {
      offsetMode = hexadecimal;
    }
    else
    {
      offsetMode = decimal;
    }
  }


  EViewMode offsetMode;
  EViewMode primaryMode;
  EViewMode secondaryMode;
  bool offsetUpperCase;
  bool primaryUpperCase;
  bool offsetVisible;
  bool lockLine;
  bool lockColumn;
  uint lineSize;
  uint columnSize;
  int  columnSpacing;
  uint separatorMarginWidth;
  uint edgeMarginWidth;
  uint leftSeparatorWidth;
  uint rightSeparatorWidth;
};




struct SDisplayCursor
{
  enum EFocusMode
  {
    stopBlinking = 0,
    hide,
    ignore,
  };

  SDisplayCursor( void )
  {
    focusMode   = hide;
    interval    = 500;
    alwaysVisible = false;
    alwaysBlockShape = false;
    thickInsertShape = true;
  }

  const char *modeStrings( uint index )
  {
    static char *str[6] = 
    {
      "stopBlinking",
      "hide",
      "ignore",
    };
    return( str[index] );
  }


  const char *focusModeString( void )
  {
    return( modeStrings( focusMode > ignore ? stopBlinking : focusMode ));
  }

  void setFocusMode( const char *str )
  {
    if( str == 0 || strcmp( str, "hide" ) == 0 )
    {
      focusMode = hide;
    }
    else if( strcmp( str, "stopBlinking" ) == 0 )
    {
      focusMode = stopBlinking;
    }
    else
    {
      focusMode = ignore;
    }
  }

  EFocusMode focusMode;
  uint interval;
  bool alwaysVisible;
  bool alwaysBlockShape;
  bool thickInsertShape;
};


struct SDisplayColor
{
  SDisplayColor( void )
  {
    useSystemColor   = false;
    textBg           = white;
    inactiveBg       = gray;
    selectBg         = kapp->selectColor;
    selectFg         = kapp->selectTextColor;
    primaryFg[0]     = black;
    primaryFg[1]     = blue;
    offsetFg         = red;
    secondaryFg      = black;
    leftSeparatorFg  = darkGreen;
    rightSeparatorFg = darkGreen;
    cursorBg         = red;
    cursorFg         = black;
  }

  bool useSystemColor;
  QColor textBg;
  QColor inactiveBg;
  QColor selectBg;
  QColor selectFg;
  QColor primaryFg[2];
  QColor offsetFg;
  QColor secondaryFg;
  QColor leftSeparatorFg;
  QColor rightSeparatorFg;
  QColor cursorBg;
  QColor cursorFg;
};


struct SDisplayFont
{
  SDisplayFont( void )
  {
    useSystemFont = false;
    localFont.setFamily( "courier" );
    localFont.setPointSize( 14 );
    localFont.setWeight( QFont::Normal );
    localFont.setItalic( false );
  }

  bool  useSystemFont;
  QFont localFont;
};
  

struct SDisplayMisc
{
  SDisplayMisc( void )
  {
    undoLevel = 100;
    autoCopyToClipboard = true;
    remeberFiles = true;
    inputSound = false;
    fatalSound = false;
  }

  uint undoLevel;
  bool autoCopyToClipboard;
  bool remeberFiles;
  bool inputSound;
  bool fatalSound;
};



struct SDisplayState
{
  SDisplayLine   line;
  SDisplayLayout layout;
  SDisplayCursor cursor;
  SDisplayColor  color;
  SDisplayFont   font;
  SDisplayMisc   misc;
};





#endif







